@@ -1,6 +1,6 @@
-import datetime
+import re
 import decimal
-import math
+import datetime
 from time import localtime
 
 from psycopg2cffi._impl.libpq import libpq, ffi
@@ -194,28 +194,12 @@ def parse_unicode(value, length, cursor):
     return value.decode(cursor._conn._py_enc) if value is not None else None
 
 
-def _parse_date(value):
-    if value is not None:
-        return datetime.date(*[int(x) for x in value.split('-')])
-
-
-def _parse_time(value, cursor):
-    """Parse the time to a datetime.time type.
-
-    The given value is in the format of `16:28:09.506488+01`
-
-    """
-    if value is not None:
-        return datetime.time(*_parse_time_to_args(value, cursor))
-
-
 def _parse_time_to_args(value, cursor):
     """Return arguemnts for datetime.time constructor
 
     The given value is in the format of `16:28:09.506488+01`
 
     """
-    microsecond = 0
     hour, minute, second = value.split(':', 2)
 
     sign = 0
@@ -230,40 +214,77 @@ def _parse_time_to_args(value, cursor):
 
     if not cursor.tzinfo_factory is None and sign:
         parts = timezone.split(':')
-        tz_min = sign * 60 * int(parts[0])
+        tz_min = 60 * int(parts[0])
         if len(parts) > 1:
             tz_min += int(parts[1])
-        if len(parts) > 2:
-            tz_min += int(int(parts[2]) / 60.0)
-        tzinfo = cursor.tzinfo_factory(tz_min)
+        if len(parts) > 2 and int(parts[2]) >= 30:
+            tz_min += 1
+        tzinfo = cursor.tzinfo_factory(sign * tz_min)
 
     if '.' in second:
-        second, microsecond = second.split('.')
-        microsecond = int(microsecond) * int(math.pow(10.0, 6.0 - len(microsecond)))
+        second, frac = second.split('.')
+        micros = int((frac + ('0' * (6 - len(frac))))[:6])
+    else:
+        micros = 0
 
-    return int(hour), int(minute), int(second), microsecond, tzinfo
+    return int(hour), int(minute), int(second), micros, tzinfo
 
 
 def parse_datetime(value, length, cursor):
     if value is None:
         return None
+    elif value == 'infinity':
+        return datetime.datetime.max
+    elif value == '-infinity':
+        return datetime.datetime.min
 
-    date, time = value.split(' ')
-    date_args = date.split('-')
-    return datetime.datetime(
-            int(date_args[0]), 
-            int(date_args[1]), 
-            int(date_args[2]), 
-            *_parse_time_to_args(time, cursor))
+    try:
+        date, time = value.split(' ')
+        date_args = date.split('-')
+        return datetime.datetime(
+                int(date_args[0]),
+                int(date_args[1]),
+                int(date_args[2]),
+                *_parse_time_to_args(time, cursor))
+    except (TypeError, ValueError):
+        if value.endswith('BC'):
+            raise ValueError('BC dates not supported')
+        raise DataError("bad datetime: '%s'" % value)
 
 
 def parse_date(value, length, cursor):
-    return _parse_date(value) if value is not None else None
+    if value is None:
+        return None
+    elif value == 'infinity':
+        return datetime.date.max
+    elif value == '-infinity':
+        return datetime.date.min
+
+    try:
+        return datetime.date(*[int(x) for x in value.split('-')])
+    except (TypeError, ValueError):
+        if value.endswith('BC'):
+            raise ValueError('BC dates not supported')
+        raise DataError("bad datetime: '%s'" % value)
 
 
 def parse_time(value, length, cursor):
-    return _parse_time(value, cursor) if value is not None else None
+    if value is None:
+        return None
 
+    try:
+        return datetime.time(*_parse_time_to_args(value, cursor))
+    except (TypeError, ValueError):
+        raise DataError("bad datetime: '%s'" % value)
+
+
+_re_interval = re.compile(r"""
+    (?:(-?\+?\d+)\sy\w+\s?)?    # year
+    (?:(-?\+?\d+)\sm\w+\s?)?    # month
+    (?:(-?\+?\d+)\sd\w+\s?)?    # day
+    (?:(-?\+?)(\d+):(\d+):(\d+) # +/- hours:mins:secs
+        (?:\.(\d+))?)?          # second fraction
+    """, re.VERBOSE)
 
 def parse_interval(value, length, cursor):
     """Typecast an interval to a datetime.timedelta instance.
@@ -275,79 +296,32 @@ def parse_interval(value, length, cursor):
     if value is None:
         return None
 
-    years = months = days = 0
-    hours = minutes = seconds = hundreths = 0.0
-    v = 0.0
-    sign = 1
-    denominator = 1.0
-    part = 0
-    skip_to_space = False
+    m = _re_interval.match(value)
+    if not m:
+        raise ValueError("failed to parse interval: '%s'" % value)
 
-    s = value
-    for c in s:
-        if skip_to_space:
-            if c == " ":
-                skip_to_space = False
-            continue
-        if c == "-":
-            sign = -1
-        elif "0" <= c <= "9":
-            v = v * 10 + ord(c) - ord("0")
-            if part == 6:
-                denominator *= 10
-        elif c == "y":
-            if part == 0:
-                years = int(v * sign)
-                skip_to_space = True
-                v = 0.0
-                sign = 1
-                part = 1
-        elif c == "m":
-            if part <= 1:
-                months = int(v * sign)
-                skip_to_space = True
-                v = 0.0
-                sign = 1
-                part = 2
-        elif c == "d":
-            if part <= 2:
-                days = int(v * sign)
-                skip_to_space = True
-                v = 0.0
-                sign = 1
-                part = 3
-        elif c == ":":
-            if part <= 3:
-                hours = v
-                v = 0.0
-                part = 4
-            elif part == 4:
-                minutes = v
-                v = 0.0
-                part = 5
-        elif c == ".":
-            if part == 5:
-                seconds = v
-                v = 0.0
-                part = 6
+    years, months, days, sign, hours, mins, secs, frac = m.groups()
 
-    if part == 4:
-        minutes = v
-    elif part == 5:
-        seconds = v
-    elif part == 6:
-        hundreths = v / denominator
+    days = int(days) if days is not None else 0
+    if months is not None:
+        days += int(months) * 30
+    if years is not None:
+        days += int(years) * 365
 
-    if sign < 0.0:
-        seconds = - (hundreths + seconds + minutes * 60 + hours * 3600)
+    if hours is not None:
+        secs = int(hours) * 3600 + int(mins) * 60 + int(secs)
+        if frac is not None:
+            micros = int((frac + ('0' * (6 - len(frac))))[:6])
+        else:
+            micros = 0
+
+        if sign == '-':
+            secs = -secs
+            micros = -micros
     else:
-        seconds += hundreths + minutes * 60 + hours * 3600
-
-    days += years * 365 + months * 30
-    micro = (seconds - math.floor(seconds)) * 1000000.0
-    seconds = int(math.floor(seconds))
-    return datetime.timedelta(days, seconds, int(micro))
+        secs = micros = 0
 
+    return datetime.timedelta(days, secs, micros)
 
 
 def Date(year, month, day):
